import inspect
import numpy as np
from monograd.tensor import Tensor
import monograd.ops as Ops 

class Module():
    def parameters(self) -> list:
        params:list = []
        for _, attr in inspect.getmembers(self):
            if isinstance(attr, Tensor) and attr.requires_grad:
                params.append(attr)
            elif isinstance(attr, Module):
                params.extend(attr.parameters())

        return params

    def zero_grad(self):
        for p in self.parameters():
            p.grad = None

    def __call__(self, *args, **kwargs):
        return self.forward(*args, **kwargs) # pyright: ignore dont know .forward exists

class Linear(Module):
    def __init__(self, in_features:int, out_features:int, bias=True):
        self.in_features = in_features
        self.out_features = out_features

        # Random Gaussian weight initialization
        std_dev = np.sqrt(2.0 / in_features)
        self.weight = Tensor(
            np.random.randn(in_features, out_features) * std_dev,
            requires_grad=True,
            name="Linear Weight"
        )

        if bias:
            self.bias = Tensor(np.zeros((out_features,)), requires_grad=True, name="Linear Bias")
        else:
            self.bias = None
    
    def forward(self, x:Tensor):
        out = x @ self.weight
        if self.bias is not None:
            out = out + self.bias
        return out

class Conv2d(Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0):
        self.stride = stride
        self.padding = padding
        self.kernel_size = kernel_size
        
        # Kaiming He Initialization
        scale = np.sqrt(2.0 / (in_channels * kernel_size * kernel_size))
        self.weight = Tensor(
            np.random.randn(out_channels, in_channels, kernel_size, kernel_size) * scale,
            requires_grad=True
        )
        self.bias = Tensor(np.zeros(out_channels), requires_grad=True)

    def forward(self, x):
        out = Ops.CONV2D.apply(x, self.weight, self.stride, self.padding)
        out = out + self.bias.reshape((1, -1, 1, 1))
        
        return out

class MaxPool2d(Module):
    def __init__(self, kernel_size, stride=None, padding=0):
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding

    def forward(self, x):
        return Ops.MAXPOOL2D.apply(x, self.kernel_size, self.stride, self.padding)



from typing import Any
import numpy as np
from monograd.utils import dbg, im2col_indices, col2im_indices

def unbroadcast(grad:np.ndarray, original_shape:tuple) -> np.ndarray:
    # Collapse leading dimensions (e.g. (32, 10) -> (10,))
    while grad.ndim > len(original_shape):
        grad = np.sum(grad, axis=0)

    # Collapse broadcasted dimensions (e.g. (4, 4) -> (4, 1) -> (4,))
    for i, dim in enumerate(original_shape):
        if dim == 1:
            grad = np.sum(grad, axis=i, keepdims=True)
    return grad

class OP():
    @staticmethod
    def forward(ctx, *args):
        raise NotImplementedError

    @staticmethod
    def backward(ctx, grad_output):
        raise NotImplementedError

    @classmethod
    def apply(cls, *args):
        # TODO: Might need to assert / transfer args into Tensor!!!
        # PERFIMPROVEMENT: do we need to have requires_grad=True in *result_tensor*
        from monograd.tensor import Context, Tensor

        # 1. create context
        ctx = Context()

        # 2. forward pass
        forward_args = [x.data if hasattr(x, "data") else x for x in args] # basically tensors
        result_data = cls.forward(ctx, *forward_args)
        result_tensor = Tensor(result_data, op=cls, parents=[x for x in args if hasattr(x, "op")]) # pyright: ignore cls is not recognized as OP

        # 3. attatch
        result_tensor.ctx = ctx
        return result_tensor

class ADD(OP):
    @staticmethod
    def forward(ctx, *args):
        x:np.ndarray = args[0]
        y:np.ndarray = args[1]
        ctx.save_for_backward(x.shape, y.shape)
        return x + y # numpy op

    @staticmethod
    def backward(ctx, grad_output):
        from monograd.tensor import Tensor
        x_shape, y_shape = ctx.saved_data
        grad_x = grad_output
        grad_y = grad_output

        # unbroadcast nparrays
        if x_shape != grad_x.shape:
            grad_x = Tensor(unbroadcast(grad_x.data, x_shape), requires_grad=False)
        if y_shape != grad_y.shape:
            grad_y = Tensor(unbroadcast(grad_y.data, y_shape), requires_grad=False)

        return grad_x, grad_y 

class SUB(OP):
    @staticmethod
    def forward(ctx, *args):
        x:np.ndarray = args[0]
        y:np.ndarray = args[1]
        ctx.save_for_backward(x.shape, y.shape)
        return x - y # numpy op

    @staticmethod
    def backward(ctx, grad_output):
        from monograd.tensor import Tensor
        x_shape, y_shape = ctx.saved_data
        grad_x = grad_output
        grad_y = -grad_output

        # unbroadcast nparrays
        if x_shape != grad_x.shape:
            grad_x = Tensor(unbroadcast(grad_x.data, x_shape), requires_grad=False)
        if y_shape != grad_y.shape:
            grad_y = Tensor(unbroadcast(grad_y.data, y_shape), requires_grad=False)

        return grad_x, grad_y 

class MUL(OP):
    @staticmethod
    def forward(ctx, *args):
        x:np.ndarray = args[0]
        y:np.ndarray = args[1]
        ctx.save_for_backward(x, y)
        return x * y  # numpy op

    @staticmethod
    def backward(ctx, grad_output):
        from monograd.tensor import Tensor
        x, y = ctx.saved_data
        grad_x = grad_output * y
        grad_y = grad_output * x

        # unbroadcast nparrays
        if x.shape != grad_x.shape:
            grad_x = Tensor(unbroadcast(grad_x.data, x.shape), requires_grad=False)
        if y.shape != grad_y.shape:
            grad_y = Tensor(unbroadcast(grad_y.data, y.shape), requires_grad=False)

        return grad_x, grad_y 

class TRANSPOSE(OP):
    @staticmethod
    def forward(ctx, *args):
        x:np.ndarray = args[0]
        order:list|tuple|None = args[1]
        ctx.save_for_backward(order)
        return np.transpose(x, order)
    
    @staticmethod
    def backward(ctx, grad_output):
        order, = ctx.saved_data
        if order is None:
            return grad_output.transpose()
        else:
            return grad_output.transpose(np.argsort(order))

class MATMUL(OP):
    @staticmethod
    def forward(ctx, *args):
        x:np.ndarray = args[0]
        y:np.ndarray = args[1]
        ctx.save_for_backward(x, y)
        return x @ y # numpy matmul

    @staticmethod
    def backward(ctx, grad_output):
        from monograd.tensor import Tensor
        x, y = ctx.saved_data
        
        x_t = Tensor(x, requires_grad=False)
        y_t = Tensor(y, requires_grad=False)
        
        # grad_x = grad_output @ y.T
        grad_x = grad_output.matmul(y_t.transpose()) # !
        # grad_y = x.T @ grad_output
        grad_y = x_t.transpose().matmul(grad_output)
        
        return grad_x, grad_y

class RELU(OP):
    @staticmethod
    def forward(ctx, *args):
        x:np.ndarray = args[0]
        ctx.save_for_backward(x)
        return np.maximum(x, 0)

    @staticmethod
    def backward(ctx, grad_output):
        from monograd.tensor import Tensor
        x, = ctx.saved_data
        a = grad_output.data * (x > 0)
        return Tensor(a)

class LEAKYRELU(OP):
    @staticmethod
    def forward(ctx, *args):
        x:np.ndarray = args[0]
        slope:float = args[1]
        ctx.save_for_backward(x, slope)
        return np.maximum(x, x * slope)

    @staticmethod
    def backward(ctx, grad_output):
        from monograd.tensor import Tensor
        x, slope = ctx.saved_data
        grad = grad_output.data.copy()
        grad[x <= 0] *= slope
        return Tensor(grad)

class SUM(OP):
    @staticmethod
    def forward(ctx, *args):
        x:np.ndarray = args[0]
        axis:int|None = args[1]
        ctx.save_for_backward(x.shape, axis)
        return np.sum(x, axis=axis)

    @staticmethod
    def backward(ctx, grad_output):
        from monograd.tensor import Tensor
        shape, axis = ctx.saved_data

        grad = grad_output.data
        if axis is not None:
            grad = np.expand_dims(grad, axis)

        return Tensor(np.ones(shape) * grad)

class RESHAPE(OP):
    @staticmethod
    def forward(ctx, *args):
        x:np.ndarray = args[0]
        shape:tuple = args[1]
        ctx.save_for_backward(x.shape)
        return x.reshape(shape)

    @staticmethod
    def backward(ctx, grad_output):
        from monograd.tensor import Tensor
        original_shape, = ctx.saved_data
        return Tensor(grad_output.data.reshape(original_shape), requires_grad=False)

class EXP(OP):
    @staticmethod
    def forward(ctx, *args):
        x:np.ndarray = args[0]
        result = np.exp(x)
        ctx.save_for_backward(result)
        return result

    @staticmethod
    def backward(ctx, grad_output):
        from monograd.tensor import Tensor
        result, = ctx.saved_data
        return Tensor(grad_output.data * result, requires_grad=False)

class LOG(OP):
    @staticmethod
    def forward(ctx, *args):
        x:np.ndarray = args[0]
        ctx.save_for_backward(x)
        return np.log(x) 

    @staticmethod
    def backward(ctx, grad_output):
        from monograd.tensor import Tensor
        x, = ctx.saved_data
        return Tensor(grad_output.data / x, requires_grad=False)

class CONV2D(OP):
    @staticmethod
    def forward(ctx, *args):
        """
        x: (N, C_in, H, W)
        w: (C_out, C_in, KH, KW)
        """
        x = args[0]
        w = args[1]
        stride = args[2]
        padding = args[3]

        N, C, H, W = x.shape
        F, C, HH, WW = w.shape # F = Filters (C_out)
        
        # 1. Math for output dimensions
        H_out = (H + 2 * padding - HH) // stride + 1
        W_out = (W + 2 * padding - WW) // stride + 1
        
        # 2. Im2Col (Turn image into a matrix)
        # Shape: (C_in * KH * KW, N * H_out * W_out)
        cols = im2col_indices(x, HH, WW, padding, stride)
        
        # 3. Flatten Weights
        # Shape: (F, C_in * KH * KW)
        w_col = w.reshape(F, -1)
        
        # 4. Matrix Multiplication (The actual convolution)
        # Shape: (F, N * H_out * W_out)
        out = w_col @ cols
        
        # 5. Reshape back to (N, F, H_out, W_out)
        out = out.reshape(F, H_out, W_out, N)
        out = out.transpose(3, 0, 1, 2)
        
        ctx.save_for_backward(x, cols, w, stride, padding)
        return out

    @staticmethod
    def backward(ctx, grad_output):
        from monograd.tensor import Tensor
        x, cols, w, stride, padding = ctx.saved_data
        
        N, C, H, W = x.shape
        F, _, HH, WW = w.shape
        
        # Reshape Gradient for Matrix Mult
        # (N, F, H_out, W_out) -> (F, N * H_out * W_out)
        grad_reshaped = grad_output.data.transpose(1, 2, 3, 0).reshape(F, -1)
        
        # Compute Weight Gradient
        # dW = grad * cols.T
        grad_w_col = grad_reshaped @ cols.T
        grad_w = grad_w_col.reshape(w.shape)
        
        # Compute Input Gradient (This needs Col2Im)
        # dX = W.T * grad
        w_reshape = w.reshape(F, -1)
        grad_cols = w_reshape.T @ grad_reshaped
        
        # Turn columns back into image
        grad_x = col2im_indices(grad_cols, x.shape, HH, WW, padding, stride)
        
        return Tensor(grad_x), Tensor(grad_w)

class MAXPOOL2D(OP):
    @staticmethod
    def forward(ctx, *args):
        x = args[0]
        kernel_size = args[1]
        stride = args[2]
        padding = args[3]

        # Default stride to kernel_size (standard for pooling)
        if stride is None: stride = kernel_size
            
        N, C, H, W = x.shape
        H_out = (H + 2 * padding - kernel_size) // stride + 1
        W_out = (W + 2 * padding - kernel_size) // stride + 1

        # shape: (C * kernel_size * kernel_size, N * H_out * W_out)
        cols = im2col_indices(x, kernel_size, kernel_size, padding, stride)
        
        # New shape: (kernel_size * kernel_size, C, N * H_out * W_out)
        cols_reshaped = cols.reshape(C, kernel_size * kernel_size, -1)
        
        out = np.max(cols_reshaped, axis=1)
        
        # Argmax gives the index of the max value (0 to k*k-1)
        argmax = np.argmax(cols_reshaped, axis=1)
        
        # Reshape Output
        out = out.reshape(C, H_out, W_out, N)
        out = out.transpose(3, 0, 1, 2)

        ctx.save_for_backward(x.shape, argmax, kernel_size, stride, padding)
        return out

    @staticmethod
    def backward(ctx, grad_output):
        from monograd.tensor import Tensor
        x_shape, argmax, kernel_size, stride, padding = ctx.saved_data
        N, C, H, W = x_shape
        
        # Flatten gradient to match the column shape
        grad = grad_output.data.transpose(1, 2, 3, 0).reshape(C, -1)
        
        # Create a zero matrix for the columns
        cols_grad = np.zeros((C, kernel_size * kernel_size, grad.shape[1]))
        
        # Route the gradient ONLY to the max indices
        rows = np.arange(C)[:, None]
        cols_indices = np.arange(grad.shape[1])[None, :]
        cols_grad[rows, argmax, cols_indices] = grad
        
        # Flatten back to im2col shape
        cols_grad = cols_grad.reshape(C * kernel_size * kernel_size, -1)
        
        # Col2Im
        grad_x = col2im_indices(cols_grad, x_shape, kernel_size, kernel_size, padding, stride)
        
        return Tensor(grad_x)

class LOADOP(OP):
    @staticmethod
    def forward(ctx, *args):
        raise TypeError

    @staticmethod
    def backward(ctx, grad_output):
        raise TypeError




from typing import List, Tuple
from monograd.utils import dbg, Device
import monograd.ops as Ops
import numpy as np 

class Context(): # Tensor context
    saved_data:List

    def __init__(self):
        self.saved_data = []

    def save_for_backward(self, *args): # MEMIMPROVEMENT: only save x.data when x is Tensor :: its runtime 4 memory payoff
        self.saved_data.extend(args)

class Tensor():
    op:type|None
    data:np.ndarray # for now we save it as ndarray but slow
    parents:Tuple|None
    device:Device
    requires_grad:bool
    parents:Tuple|None
    name:str|None
    ctx:Context|None

    def __init__(self, data:List|np.ndarray|int|float, op:type|None=None, parents:Tuple|None = None, requires_grad:bool=True, _dtype=np.float32, name=None):
        self.name:str|None = name 
        self.op = op
        assert not self.op or issubclass(self.op, Ops.OP)

        # data & grad
        if isinstance(data, (List, Tuple, int, float, np.number)):
            self.data = np.array(data, dtype=_dtype)
        elif isinstance(data, np.ndarray): self.data = data
        elif isinstance(data, Tensor): dbg(f"[TENSOR] Got type {type(data)} as *data* in Tensor")
        assert isinstance(self.data, np.ndarray)
        self.grad:Tensor|None = None
        self.requires_grad = requires_grad

        if not parents: self.parents = tuple()
        else: self.parents = tuple(parents)
        self.shape = self.data.shape
        self.strides = self.data.strides
        self.offset = 0
        self._dtype = _dtype

        self.device = Device.CPU
        self.ctx = None

    def backward(self):
        graph: List[Tensor] = _toposort(self)
        if not self.grad: 
            self.grad = Tensor(np.ones(self.shape, dtype=self._dtype), requires_grad=False)

        for node in graph:
            # check if we can actually call op.backward
            if not node.op or isinstance(node.op, Ops.LOADOP):
                continue

            grads = node.op.backward(node.ctx, node.grad)
            if not isinstance(grads, (tuple, list)): grads = (grads,)

            if not node.parents:
                continue

            for parent, grad in zip(node.parents, grads):
                if not parent.requires_grad:
                    continue

                if not parent.grad: parent.grad = grad
                else: parent.grad += grad

    def relu(self):
        return Ops.RELU.apply(self)

    def leaky_relu(self, slope=0.01):
        return Ops.LEAKYRELU.apply(self, slope)

    def sum(self, axis=None):
        return Ops.SUM.apply(self, axis)

    def matmul(self, other):
        if not isinstance(other, Tensor):
            other = Tensor(other)
        return Ops.MATMUL.apply(self, other)

    def transpose(self, order=None):
        return Ops.TRANSPOSE.apply(self, order)

    def reshape(self, shape):
        return Ops.RESHAPE.apply(self, shape)

    def exp(self):
        return Ops.EXP.apply(self)

    def log(self):
        return Ops.LOG.apply(self)

    @property
    def T(self):
        return self.transpose()
      
    def toposort(self): # REMOVE: used rn for testing 
        return _toposort(self)

    def __neg__(self):
        return self * -1

    def __rsub__(self, other):
        if not isinstance(other, Tensor):
            other = Tensor(other)
        return other + (-self)

    def __add__(self, other):
        if not isinstance(other, Tensor):
            other = Tensor(other)
        return Ops.ADD.apply(self, other)

    def __sub__(self, other):
        if not isinstance(other, Tensor):
            other = Tensor(other)
        return Ops.SUB.apply(self, other)

    def __mul__(self, other):
        if not isinstance(other, Tensor):
            other = Tensor(other)
        return Ops.MUL.apply(self, other)
    
    def __matmul__(self, other):
        return self.matmul(other)

    def __repr__(self):
        return f"<Tensor name={self.name} op={self.op} data={self.data} device={self.device} requires_grad={self.requires_grad}>"

def _toposort(leaf:Tensor) -> List[Tensor]: # TODO: make tests
    # topological sort algo to order DAG
    visited:set = set()
    stack:List = []

    def dfs(t:Tensor):
        visited.add(t)
        parents = t.parents
        if parents:
            for parent in parents:
                if parent not in visited: dfs(parent)
        stack.append(t)

    dfs(leaf)
    return stack[::-1]




import numpy as np
import os
import gzip
import urllib.request

def mnist():
    url_base = "https://storage.googleapis.com/cvdf-datasets/mnist/"
    file_names = {
        "x_train": "train-images-idx3-ubyte.gz",
        "y_train": "train-labels-idx1-ubyte.gz",
        "x_test": "t10k-images-idx3-ubyte.gz",
        "y_test": "t10k-labels-idx1-ubyte.gz"
    }
    
    data = {}
    save_dir = "./data"
    if not os.path.exists(save_dir): os.makedirs(save_dir)
        
    for key, filename in file_names.items():
        filepath = os.path.join(save_dir, filename)
        if not os.path.exists(filepath):
            urllib.request.urlretrieve(url_base + filename, filepath)
            
        with gzip.open(filepath, 'rb') as f:
            if "images" in filename:
                # Skip 16-byte header
                arr = np.frombuffer(f.read(), np.uint8, offset=16)
                data[key] = arr.reshape(-1, 28, 28).astype(np.float32) / 255.0
            else:
                # Skip 8-byte header
                data[key] = np.frombuffer(f.read(), np.uint8, offset=8)
                
    return data["x_train"], data["y_train"], data["x_test"], data["y_test"]

# if __name__ == "__main__":
#     xt, yt, xv, yv = mnist()
#     print(f"X Train: {xt.shape}")  # Should be (60000, 28, 28)
#     print(f"Y Train: {yt.shape}")  # Should be (60000,)



# Unfortunatly no good way to mytype this cuz of redundancy imports
import numpy as np

class Optimizer():
    def __init__(self, params, lr):
        self.params:list = list(params) # params can be Tensor or nn.Module?
        self.lr:float = lr
    
    def step(self):
        for tensor in self.params:
            tensor.data -= self.lr * tensor.grad.data

    def zero_grad(self):
        for tensor in self.params:
            tensor.grad = None

class SGD(Optimizer):
    def __init__(self, params, lr=0.1, momentum=0.0):
        super().__init__(params, lr)
        self.momentum = momentum
        self.velocities = []

        for p in self.params:
            self.velocities.append(np.zeros_like(p.data))

    def step(self):
        for i, p in enumerate(self.params):
            if p.grad is None:
                continue
            # v = (momentum * v) + grad
            self.velocities[i] = (self.momentum * self.velocities[i]) + p.grad.data
            # w = w - (lr * v)
            p.data -= self.lr * self.velocities[i]

class Adam(Optimizer):
    pass



from enum import Enum
import numpy as np
import os

DEBUG_MODE = os.environ.get("DEBUG", "").lower() in ("true", "1", "t")
def dbg(*args, **kwargs):
    if DEBUG_MODE:
        print(*args, **kwargs)

class Device(Enum):
    CPU = 1
    GPU = 2
    METAL = 3

def get_im2col_indices(x_shape, field_height, field_width, padding=1, stride=1):
    N, C, H, W = x_shape
    assert (H + 2 * padding - field_height) % stride == 0
    assert (W + 2 * padding - field_width) % stride == 0
    
    out_height = (H + 2 * padding - field_height) // stride + 1
    out_width = (W + 2 * padding - field_width) // stride + 1

    # Creating vectors for rows and columns
    i0 = np.repeat(np.arange(field_height), field_width)
    i0 = np.tile(i0, C)
    i1 = stride * np.repeat(np.arange(out_height), out_width)
    j0 = np.tile(np.arange(field_width), field_height * C)
    j1 = stride * np.tile(np.arange(out_width), out_height)
    
    # Broadcast to create indices
    i = i0.reshape(-1, 1) + i1.reshape(1, -1)
    j = j0.reshape(-1, 1) + j1.reshape(1, -1)
    k = np.repeat(np.arange(C), field_height * field_width).reshape(-1, 1)

    return (k, i, j)

def im2col_indices(x, field_height, field_width, padding=1, stride=1):
    p = padding
    
    if isinstance(p, tuple) or isinstance(p, list):
        p = p[0]
        
    x_padded = np.pad(x, ((0, 0), (0, 0), (p, p), (p, p)), mode='constant')
    k, i, j = get_im2col_indices(x.shape, field_height, field_width, padding, stride)

    cols = x_padded[:, k, i, j]
    C = x.shape[1]
    cols = cols.transpose(1, 2, 0).reshape(field_height * field_width * C, -1)
    return cols

def col2im_indices(cols, x_shape, field_height, field_width, padding=1, stride=1):
    N, C, H, W = x_shape
    H_padded, W_padded = H + 2 * padding, W + 2 * padding
    x_padded = np.zeros((N, C, H_padded, W_padded), dtype=cols.dtype)
    
    k, i, j = get_im2col_indices(x_shape, field_height, field_width, padding, stride)
    
    cols_reshaped = cols.reshape(C * field_height * field_width, -1, N)
    cols_reshaped = cols_reshaped.transpose(2, 0, 1)
    
    # Using np.add.at for atomic addition (handling overlapping patches correctly)
    np.add.at(x_padded, (slice(None), k, i, j), cols_reshaped) # pyright: ignore
    
    if padding == 0:
        return x_padded
    return x_padded[:, :, padding:-padding, padding:-padding]
